# -*- coding: utf-8 -*-

import zope.interface
import zope.component
import zope.security
import zope.security.proxy
import zope.traversing.adapters

import parser
import re

import interfaces
import types

_marker = object()

class ExpressionTranslation(object):
    zope.interface.implements(interfaces.IExpressionTranslation)

    re_pragma = re.compile(r'^\s*(?P<pragma>[a-z]+):\s*')
    re_interpolation = re.compile(r'(?P<prefix>[^\\]\$|^\$){((?P<expression>.*)})?')

    def name(self, string):
        return string
    
    def search(self, string):
        """
        We need to implement a ``validate``-method. Let's define that an
        expression is valid if it contains an odd number of
        characters.
        
          >>> class MockExpressionTranslation(ExpressionTranslation):
          ...     def validate(self, string):
          ...         if len(string) % 2 == 0: raise SyntaxError()

          >>> search = MockExpressionTranslation().search
          >>> search("a")
          'a'
          >>> search("ab")
          'a'
          >>> search("abc")
          'abc'
          
        """

        left = 0
        right = left + 1

        current = None

        while right <= len(string):
            expression = string[left:right]

            try:
                e = self.validate(expression)
                current = expression
            except SyntaxError, e:
                if right == len(string):
                    if current is not None:
                        return current

                    raise e

            right += 1

        return current

    def declaration(self, string):
        """
          >>> declaration = ExpressionTranslation().declaration

        Single variable:

          >>> declaration("variable")
          declaration('variable',)

        Multiple variables:

          >>> declaration("variable1, variable2")
          declaration('variable1', 'variable2')
        """

        variables = []
        for var in string.split(', '):
            var = var.strip()

            if var in ('repeat',):
                raise ValueError, "Invalid variable name '%s' (reserved)." % variable

            if var.startswith('_') and not var.startswith('_tmp'):
                raise ValueError(
                    "Invalid variable name '%s' (starts with an underscore)." % var)

            variables.append(var)

        return types.declaration(variables)

    def mapping(self, string):
        """
          >>> mapping = ExpressionTranslation().mapping
          
          >>> mapping("abc def")
          mapping(('abc', 'def'),)

          >>> mapping("abc")
          mapping(('abc', None),)

          >>> mapping("abc; def ghi")
          mapping(('abc', None), ('def', 'ghi'))

        """

        defs = string.split(';')
        mappings = []
        for d in defs:
            d = d.strip()
            while '  ' in d:
                d = d.replace('  ', ' ')

            parts = d.split(' ')
            if len(parts) == 1:
                mappings.append((d, None))
            elif len(parts) == 2:
                mappings.append((parts[0], parts[1]))
            else:
                raise ValueError, "Invalid mapping (%s)." % string

        return types.mapping(mappings)

    def definitions(self, string):
        """
        
        >>> class MockExpressionTranslation(ExpressionTranslation):
        ...     def expression(self, string):
        ...         return types.value(string.strip())
        
        >>> definitions = MockExpressionTranslation().definitions
        
        Single define:
        
        >>> definitions("variable expression")
        definitions((declaration('variable',), value('expression')),)
        
        Multiple defines:
        
        >>> definitions("variable1 expression1; variable2 expression2")
        definitions((declaration('variable1',), value('expression1')),
                    (declaration('variable2',), value('expression2')))
        
        Tuple define:
        
        >>> definitions("(variable1, variable2) (expression1, expression2)")
        definitions((declaration('variable1', 'variable2'),
                    value('(expression1, expression2)')),)
        
        A define clause that ends in a semicolon:
        
        >>> definitions("variable expression;")
        definitions((declaration('variable',), value('expression')),)
        
        A define clause with a trivial expression (we do allow this):
        
        >>> definitions("variable")
        definitions((declaration('variable',), None),)
        
        A proper define clause following one with a trivial expression:
        
        >>> definitions("variable1 expression; variable2")
        definitions((declaration('variable1',), value('expression')),
                    (declaration('variable2',), None))

        """

        string = string.replace('\n', '').strip()

        defines = []

        i = 0
        while i < len(string):
            while string[i] == ' ':
                i += 1

            # get variable definition
            if string[i] == '(':
                j = string.find(')', i+1)
                if j == -1:
                    raise ValueError, "Invalid variable tuple definition (%s)." % string
                var = self.declaration(string[i+1:j])
                j += 1
            else:
                j = string.find(' ', i + 1)
                if j == -1:
                    var = self.declaration(string[i:])
                    j = len(string)
                else:
                    var = self.declaration(string[i:j])

            # get expression
            i = j
            while j < len(string):
                j = string.find(';', j+1)
                if j == -1:
                    j = len(string)

                try:
                    expr = self.expression(string[i:j])
                except SyntaxError, e:
                    if j < len(string):
                        continue
                        
                    raise e
                break
            else:
                expr = None

            defines.append((var, expr))

            i = j + 1

        return types.definitions(defines)

    def definition(self, string):
        defs = self.definitions(string)
        if len(defs) != 1:
            raise ValueError, "Multiple definitions not allowed."

        return defs[0]

    def output(self, string):
        """
        >>> class MockExpressionTranslation(ExpressionTranslation):
        ...     def validate(self, string):
        ...         return True
        ...
        ...     def translate(self, string):
        ...         return types.value(string)

        >>> output = MockExpressionTranslation().output

        >>> output("context/title")
        escape(value('context/title'),)

        >>> output("context/pretty_title_or_id|context/title")
        escape(value('context/pretty_title_or_id'), value('context/title'))

        >>> output("structure context/title")
        value('context/title')
        
        """
        
        if string.startswith('structure '):
            return self.expression(string[len('structure'):])
        
        expression = self.expression(string)

        if isinstance(expression, types.parts):
            return types.escape(expression)

        return types.escape((expression,))
            
    def expression(self, string):
        """We need to implement the ``validate`` and
        ``translate``-methods. Let's define that an expression is
        valid if it contains an odd number of characters.
        
        >>> class MockExpressionTranslation(ExpressionTranslation):
        ...     def validate(self, string):
        ...         return True
        ...
        ...     def translate(self, string):
        ...         return types.value(string)

        >>> expression = MockExpressionTranslation().expression

        >>> expression('a')
        value('a')

        >>> expression('a|b')
        parts(value('a'), value('b'))
    
        """

        if string is None:
            import pdb; pdb.set_trace()
            
        string = string.replace('\n', '').strip()

        if not string:
            return types.parts()

        parts = []

        # default translator is ``self``
        translator = self

        i = j = 0
        while i < len(string):
            if translator is self:
                match = self.re_pragma.match(string[i:])
                if match is not None:
                    pragma = match.group('pragma')

                    translator = \
                        zope.component.queryUtility(
                            interfaces.IExpressionTranslation, name=pragma) or \
                        zope.component.queryAdapter(
                            self, interfaces.IExpressionTranslation, name=pragma) or \
                        self

                    if translator is not self:
                        i += match.end()
                        continue
            
            j = string.find('|', j + 1)
            if j == -1:
                j = len(string)

            expr = string[i:j].lstrip()

            try:
                translator.validate(expr)
            except Exception, e:
                if j < len(string):
                    continue

                raise e

            value = translator.translate(expr)
            parts.append(value)
            translator = self
            
            i = j + 1

        if len(parts) == 1:
            return parts[0]

        return types.parts(parts)

    def interpolate(self, string):
        """Search for an interpolation and return a match.

        >>> class MockExpressionTranslation(ExpressionTranslation):
        ...     def validate(self, string):
        ...         if '}' in string: raise SyntaxError
        ...
        ...     def translate(self, string):
        ...         return types.value(string)

        >>> interpolate = MockExpressionTranslation().interpolate
        
        >>> interpolate('${abc}').group('expression')
        'abc'

        >>> interpolate('abc${def}').group('expression')
        'def'

        >>> interpolate('abc${def}ghi${jkl}').group('expression')
        'def'

        >>> interpolate('${abc')
        Traceback (most recent call last):
          ...
        SyntaxError: Interpolation expressions must of the form ${<expression>} (${abc)
        
        """

        m = self.re_interpolation.search(string)
        if m is None:
            return None

        expression = m.group('expression')

        if expression:
            left = m.start()+len(m.group('prefix'))
            exp = self.search(string[left+1:])
            right = left+2+len(exp)
            m = self.re_interpolation.search(string[:right])
            
        if expression is None or m is None:
            raise SyntaxError(
                "Interpolation expressions must of the "
                "form ${<expression>} (%s)" % string)

        return m

class PythonTranslation(ExpressionTranslation):
    def validate(self, string):
        """We use the ``parser`` module to determine if
        an expression is a valid python expression."""
        
        parser.expr(string.encode('utf-8'))

    def translate(self, string):
        if isinstance(string, unicode):
            string = string.encode('utf-8')

        return types.value(string)
            
class StringTranslation(ExpressionTranslation):
    zope.component.adapts(interfaces.IExpressionTranslation)

    def __init__(self, translator):
        self.translator = translator

    def validate(self, string):
        self.interpolate(string)

    def translate(self, string):
        return types.join(self.split(string))
        
    def split(self, string):
        """Split up an interpolation string expression into parts that
        are either unicode strings or ``value``-tuples.

        >>> class MockTranslation(ExpressionTranslation):
        ...     def validate(self, string):
        ...         if '}' in string: raise SyntaxError
        ...
        ...     def translate(self, string):
        ...         return types.value(string)
        
        >>> class MockStringTranslation(StringTranslation):
        ...     pass
        
        >>> split = MockStringTranslation(MockTranslation()).split

        >>> split("${abc}")
        (value('abc'),)

        >>> split("abc${def}")
        ('abc', value('def'))

        >>> split("${def}abc")
        (value('def'), 'abc')

        >>> split("abc${def}ghi")
        ('abc', value('def'), 'ghi')

        >>> split("abc${def}ghi${jkl}")
        ('abc', value('def'), 'ghi', value('jkl'))

        >>> split("abc${def | ghi}")
        ('abc', parts(value('def '), value('ghi')))

        >>> print split("abc${La Peña}")
        ('abc', value('La Pe\\xc3\\xb1a'))
        
        """

        m = self.translator.interpolate(string)
        if m is None:
            return (string,)

        parts = []
        
        start = m.start()
        if start > 0:
            text = string[:m.start()+1]
            parts.append(text)

        expression = m.group('expression')
        parts.append(self.translator.expression(expression))

        rest = string[m.end():]
        if len(rest):
            parts.extend(self.split(rest))

        return tuple(parts)

class PathTranslation(ExpressionTranslation):
    path_regex = re.compile(r'^((nocall|not):\s*)*([A-Za-z_]+)(/[A-Za-z_@-]+)*$')

    @classmethod
    def traverse(cls, base, request, call, *path_items):
        """See ``zope.app.pagetemplate.engine``."""

        _callable = callable(base)
        
        for i in range(len(path_items)):
            name = path_items[i]

            next = getattr(base, name, _marker)
            if next is not _marker:
                base = next()
            else:
                # special-case dicts for performance reasons        
                if getattr(base, '__class__', None) == dict:
                    base = base[name]
                else:
                    base = zope.traversing.adapters.traversePathElement(
                        base, name, path_items[i+1:], request=request)

            _callable = callable(base)

            if not isinstance(base, (basestring, tuple, list)):
                base = zope.security.proxy.ProxyFactory(base)

        if call and _callable:
            base = base()
                
        return base

    def validate(self, string):
        if not self.path_regex.match(string):
            raise SyntaxError("Not a valid path-expression.")

    def translate(self, string):
        """
            >>> translate = PathTranslation().translate
            >>> translate("a/b")
            value("_path(a, request, True, 'b')")

            >>> translate("context/@@view")
            value("_path(context, request, True, '@@view')")

            >>> translate("nocall: context/@@view")
            value("_path(context, request, False, '@@view')")

            >>> translate("not: context/@@view")
            value("not(_path(context, request, True, '@@view'))")

        """

        nocall = False
        negate = False

        while string:
            m = self.re_pragma.match(string)
            if m is None:
                break

            string = string[m.end():]
            pragma = m.group('pragma').lower()

            if pragma == 'nocall':
                nocall = True
            elif pragma == 'not':
                negate = True
            else:
                raise ValueError("Invalid pragma: %s" % pragma)

        parts = string.split('/')

        # map 'nothing' to 'None'
        parts = map(lambda part: part == 'nothing' and 'None' or part, parts)
        
        base = parts[0]
        components = [repr(part) for part in parts[1:]]

        if not components:
            components = ()

        value = types.value(
            '_path(%s, request, %s, %s)' % (base, not nocall, ', '.join(components)))

        if negate:
            value = types.value('not(%s)' % value)

        return value



import os
import unittest
import time
import sys

reload(sys)
sys.setdefaultencoding('utf-8')

import zope.component.testing
import zope.configuration.xmlconfig

import zope.pagetemplate.pagetemplatefile
import z3c.pt

def benchmark(title):
    def decorator(f):
        def wrapper(*args):
            print "==========================\n %s\n==========================" % title
            return f(*args)
        return wrapper
    return decorator

def timing(func, *args, **kwargs):
    t1 = t2 = time.time()
    i = 0
    while t2 - t1 < 3:
        func(*args, **kwargs)
        i += 1
        t2 = time.time()
    return 100*(t2-t1)/i
           
class BenchmarkTestCase(unittest.TestCase):
    helloworld_z3c = z3c.pt.PageTemplate("""\
    <div xmlns="http://www.w3.org/1999/xhtml">
    Hello World!
    </div>""")

    helloworld_zope = zope.pagetemplate.pagetemplate.PageTemplate()
    helloworld_zope.pt_edit("""\
    <div xmlns="http://www.w3.org/1999/xhtml">
    Hello World!
    </div>""", 'text/xhtml')
    
    bigtable_python_z3c = z3c.pt.PageTemplate("""\
    <table xmlns="http://www.w3.org/1999/xhtml"
    xmlns:tal="http://xml.zope.org/namespaces/tal">
    <tr tal:repeat="row table">
    <td tal:repeat="c row.values()">
    <span tal:define="d c + 1"
    tal:attributes="class 'column-' + str(d)"
    tal:content="d" />
    </td>
    </tr>
    </table>""")

    bigtable_path_z3c = z3c.pt.PageTemplate("""\
    <table xmlns="http://www.w3.org/1999/xhtml"
    xmlns:tal="http://xml.zope.org/namespaces/tal"
    tal:default-expression="path">
    <tr tal:repeat="row table">
    <td tal:repeat="c row/values">
    <span tal:define="d python: c + 1"
    tal:attributes="class string:column-${d}"
    tal:content="d" />
    </td>
    </tr>
    </table>""")

    bigtable_python_zope = zope.pagetemplate.pagetemplate.PageTemplate()
    bigtable_python_zope.pt_edit("""\
    <table xmlns="http://www.w3.org/1999/xhtml"
    xmlns:tal="http://xml.zope.org/namespaces/tal">
    <tr tal:repeat="row python: options['table']">
    <td tal:repeat="c python: row.values()">
    <span tal:define="d python: c + 1"
    tal:attributes="class python:'column-'+str(d)"
    tal:content="d" />
    </td>
    </tr>
    </table>""", 'text/xhtml')

    bigtable_path_zope = zope.pagetemplate.pagetemplate.PageTemplate()
    bigtable_path_zope.pt_edit("""\
    <table xmlns="http://www.w3.org/1999/xhtml"
    xmlns:tal="http://xml.zope.org/namespaces/tal">
    <tr tal:repeat="row options/table">
    <td tal:repeat="c row/values">
    <span tal:define="d python: c + 1"
    tal:attributes="class string:column-${d}"
    tal:content="d" />
    </td>
    </tr>
    </table>""", 'text/xhtml')

    def setUp(suite):
        zope.component.testing.setUp(suite)
        zope.configuration.xmlconfig.XMLConfig('configure.zcml', z3c.pt)()

    def tearDown(suite):
        zope.component.testing.tearDown(suite)

    @benchmark(u"Hello World")
    def testHelloWorld(self):
        t_z3c = timing(self.helloworld_z3c)
        t_zope = timing(self.helloworld_zope)

        print "z3c.pt:            %.2f" % t_z3c
        print "zope.pagetemplate: %.2f" % t_zope
        print "                   %.2fX" % (t_zope/t_z3c)

    @benchmark(u"Big table (python)")
    def testBigTablePython(self):
        table = [dict(a=1,b=2,c=3,d=4,e=5,f=6,g=7,h=8,i=9,j=10) \
                 for x in range(1000)]

        t_z3c = timing(self.bigtable_python_z3c, table=table)
        t_zope = timing(self.bigtable_python_zope, table=table)

        print "z3c.pt:            %.2f" % t_z3c
        print "zope.pagetemplate: %.2f" % t_zope
        print "                   %.2fX" % (t_zope/t_z3c)

    @benchmark(u"Big table (path)")
    def testBigTablePath(self):
        table = [dict(a=1,b=2,c=3,d=4,e=5,f=6,g=7,h=8,i=9,j=10) \
                 for x in range(1000)]

        t_z3c = timing(self.bigtable_path_z3c, table=table, request=object())
        t_zope = timing(self.bigtable_path_zope, table=table)

        print "z3c.pt:            %.2f" % t_z3c
        print "zope.pagetemplate: %.2f" % t_zope
        print "                   %.2fX" % (t_zope/t_z3c)

    @benchmark(u"Compilation")
    def testCompilation(self):
        table = [dict(a=1,b=2,c=3,d=4,e=5,f=6,g=7,h=8,i=9,j=10) \
                 for x in range(1000)]

        t_z3c = timing(self.bigtable_python_z3c.cook, ['table'])
        t_zope = timing(self.bigtable_python_zope._cook)

        print "z3c.pt:            %.2f" % t_z3c
        print "zope.pagetemplate: %.2f" % t_zope
        print "                   %.2fX" % (t_zope/t_z3c)

def test_suite():
    return unittest.makeSuite(BenchmarkTestCase)

if __name__ == "__main__":
    unittest.main(defaultTest="test_suite")




from pagetemplate import PageTemplate
from pagetemplate import PageTemplateFile
from pagetemplate import ViewPageTemplate
from pagetemplate import ViewPageTemplateFile

from texttemplate import TextTemplate
from texttemplate import TextTemplateFile
from texttemplate import ViewTextTemplate
from texttemplate import ViewTextTemplateFile



import os
import sys

import translation
import template

class PageTemplate(template.BaseTemplate):
    @property
    def translate(self):
        return translation.translate_xml

class PageTemplateFile(template.BaseTemplateFile):
    @property
    def translate(self):
        return translation.translate_xml

class ViewPageTemplate(property):
    def __init__(self, body, **kwargs):
        self.template = PageTemplate(body, **kwargs)
        property.__init__(self, self.render)

    def render(self, view):
        def template(**kwargs):
            return self.template.render(view=view,
                                        context=view.context,
                                        request=view.request,
                                        options=kwargs)
        return template        
    
class ViewPageTemplateFile(ViewPageTemplate):
    def __init__(self, filename, **kwargs):
        self.template = PageTemplateFile(filename, **kwargs)
        property.__init__(self, self.render)



import sys
import re
import logging

# check if we're able to coerce unicode to str
try:
    str(u'La Pe\xf1a')
    unicode_required_flag = False
except UnicodeEncodeError:
    unicode_required_flag = True
    log = logging.getLogger('z3c.pt')
    log.info("Default system encoding is set to '%s'; "
             "the template engine will perform better if "
             "an encoding that coerces gracefully to "
             "unicode is used ('utf-8' recommended)." % sys.getdefaultencoding())

def handler(key=None):
    def decorate(f):
        def g(node):
            if key is None:
                return f(node, None)
            return f(node, node.get(key))
        g.__ns__ = key
        return g
    return decorate

s_counter = 0

class scope(list):
    def __init__(self, *args):
        global s_counter
        self.hash = s_counter
        s_counter += 1

    def __hash__(self):
        return self.hash

class repeatitem(object):
    def __init__(self, iterator, length):
        self.length = length
        self.iterator = iterator
        
    @property
    def index(self):
        try:
            length = len(self.iterator)
        except TypeError:
            length = self.iterator.__length_hint__()
        except:
            raise TypeError("Unable to determine length.")

        try:
            return self.length - length - 1
        except TypeError:
            return None
            
    @property
    def start(self):
        return self.index == 0

    @property
    def end(self):
        return self.index == self.length - 1

    def number(self):
        return self.index + 1

    def odd(self):
        return bool(self.index % 2)

    def even(self):
        return not self.odd()

class repeatdict(dict):
    def insert(self, key, iterable):
        length = len(iterable)
        iterator = iterable.__iter__()
        self[key] = (iterator, length)
        return iterator
        
    def __getitem__(self, key):
        value, length = dict.__getitem__(self, key)

        if not isinstance(value, repeatitem):
            value = repeatitem(value, length)
            self[key] = (value, length)
            
        return value



# -*- coding: utf-8 -*-

from utils import unicode_required_flag
from cgi import escape

import types

class Assign(object):
    """
    >>> from z3c.pt.generation import CodeIO; stream = CodeIO()
    >>> from z3c.pt.testing import pyexp

    We'll define some values for use in the tests.
    
    >>> one = types.value("1")
    >>> bad_float = types.value("float('abc')")
    >>> abc = types.value("'abc'")
    >>> ghi = types.value("'ghi'")
    >>> utf8_encoded = types.value("'La Peña'")
    >>> exclamation = types.value("'!'")
        
    Simple value assignment:
    
    >>> assign = Assign(one)
    >>> assign.begin(stream, 'a')
    >>> exec stream.getvalue()
    >>> a == 1
    True
    >>> assign.end(stream)
    
    Try-except parts (bad, good):
    
    >>> assign = Assign(types.parts((bad_float, one)))
    >>> assign.begin(stream, 'b')
    >>> exec stream.getvalue()
    >>> b == 1
    True
    >>> assign.end(stream)
    
    Try-except parts (good, bad):
    
    >>> assign = Assign(types.parts((one, bad_float)))
    >>> assign.begin(stream, 'b')
    >>> exec stream.getvalue()
    >>> b == 1
    True
    >>> assign.end(stream)
    
    Join:

    >>> assign = Assign(types.join((abc, ghi)))
    >>> assign.begin(stream, 'b')
    >>> exec stream.getvalue()
    >>> b == 'abcghi'
    True
    >>> assign.end(stream)

    Join with try-except parts:
    
    >>> assign = Assign(types.join((types.parts((bad_float, abc, ghi)), ghi)))
    >>> assign.begin(stream, 'b')
    >>> exec stream.getvalue()
    >>> b == 'abcghi'
    True
    >>> assign.end(stream)

    UTF-8 coercing:

    >>> assign = Assign(types.join((utf8_encoded, exclamation)))
    >>> assign.begin(stream, 'b')
    >>> exec stream.getvalue()
    >>> b == 'La Peña!'
    True
    >>> assign.end(stream)

    UTF-8 coercing with unicode:
    
    >>> assign = Assign(types.join((utf8_encoded, u"!")))
    >>> assign.begin(stream, 'b')
    >>> exec stream.getvalue()
    >>> b == 'La Peña!'
    True
    >>> assign.end(stream)

    """

    def __init__(self, parts, variable=None):
        if not isinstance(parts, types.parts):
            parts = types.parts((parts,))
        
        self.parts = parts
        self.variable = variable
        
    def begin(self, stream, variable=None):
        """First n - 1 expressions must be try-except wrapped."""

        variable = variable or self.variable

        for value in self.parts[:-1]:
            stream.write("try:")
            stream.indent()

            self._assign(variable, value, stream)
            
            stream.outdent()
            stream.write("except Exception, e:")
            stream.indent()

        value = self.parts[-1]
        self._assign(variable, value, stream)
        
        stream.outdent(len(self.parts)-1)

    def _assign(self, variable, value, stream):
        stream.annotate(value)
        
        if isinstance(value, types.value):
            stream.write("%s = %s" % (variable, value))
        elif isinstance(value, types.join):
            parts = []
            _v_count = 0
            
            for part in value:
                if isinstance(part, (types.parts, types.join)):
                    _v = stream.save()
                    assign = Assign(part, _v)
                    assign.begin(stream)
                    assign.end(stream)
                    _v_count +=1
                    parts.append(_v)
                elif isinstance(part, types.value):
                    parts.append(part)
                elif isinstance(part, unicode):
                    parts.append(repr(part.encode('utf-8')))
                elif isinstance(part, str):
                    parts.append(repr(part))
                else:
                    raise ValueError("Not able to handle %s" % type(part))
                    
            format = "%s"*len(parts)

            stream.write("%s = '%s' %% (%s)" % (variable, format, ",".join(parts)))
            
            for i in range(_v_count):
                stream.restore()
        
    def end(self, stream):
        pass
        
class Define(object):
    """
      >>> from z3c.pt.generation import CodeIO; stream = CodeIO()
      >>> from z3c.pt.testing import pyexp
      
    Variable scope:

      >>> define = Define("a", pyexp("b"))
      >>> b = object()
      >>> define.begin(stream)
      >>> exec stream.getvalue()
      >>> a is b
      True
      >>> del a
      >>> define.end(stream)
      >>> exec stream.getvalue()
      >>> a
      Traceback (most recent call last):
          ...
      NameError: name 'a' is not defined
      >>> b is not None
      True

    Multiple defines:

      >>> stream = CodeIO()
      >>> define1 = Define("a", pyexp("b"))
      >>> define2 = Define("c", pyexp("d"))
      >>> d = object()
      >>> define1.begin(stream)
      >>> define2.begin(stream)
      >>> exec stream.getvalue()
      >>> a is b and c is d
      True
      >>> define2.end(stream)
      >>> define1.end(stream)
      >>> del a; del c
      >>> stream.scope[-1].remove('a'); stream.scope[-1].remove('c')
      >>> exec stream.getvalue()
      >>> a
      Traceback (most recent call last):
          ...
      NameError: name 'a' is not defined
      >>> c
      Traceback (most recent call last):
          ...
      NameError: name 'c' is not defined
      >>> b is not None and d is not None
      True

    Tuple assignments:

      >>> stream = CodeIO()
      >>> define = Define(['e', 'f'], pyexp("[1, 2]"))
      >>> define.begin(stream)
      >>> exec stream.getvalue()
      >>> e == 1 and f == 2
      True
      >>> define.end(stream)

    Verify scope is preserved on tuple assignment:

      >>> stream = CodeIO()
      >>> e = None; f = None
      >>> stream.scope[-1].add('e'); stream.scope[-1].add('f')
      >>> stream.scope.append(set())
      >>> define.begin(stream)
      >>> define.end(stream)
      >>> exec stream.getvalue()
      >>> e is None and f is None
      True

    Using semicolons in expressions within a define:

      >>> stream = CodeIO()
      >>> define = Define("a", pyexp("';'"))
      >>> define.begin(stream)
      >>> exec stream.getvalue()
      >>> a
      ';'
      >>> define.end(stream)

    Scope:

      >>> stream = CodeIO()
      >>> a = 1
      >>> stream.scope[-1].add('a')
      >>> stream.scope.append(set())
      >>> define = Define("a", pyexp("2"))
      >>> define.begin(stream)
      >>> define.end(stream)
      >>> exec stream.getvalue()
      >>> a
      1
    
    """
    def __init__(self, definition, expression):
        if not isinstance(definition, (list, tuple)):
            definition = (definition,)

        if len(definition) == 1:
            variable = definition[0]
        else:
            variable = u"(%s,)" % ", ".join(definition)

        self.assign = Assign(expression, variable)        
        self.definitions = definition
        
    def begin(self, stream):
        # save local variables already in in scope
        for var in self.definitions:
            temp = stream.save()

            # If we didn't set the variable in this scope already
            if var not in stream.scope[-1]:

                # we'll check if it's set in one of the older scopes
                for scope in stream.scope[:-1]:
                    if var in scope:
                        # in which case we back it up
                        stream.write('%s = %s' % (temp, var))

                stream.scope[-1].add(var)
                   
        self.assign.begin(stream)

    def end(self, stream):
        self.assign.end(stream)

        # back come the variables that were already in scope in the
        # first place
        for var in reversed(self.definitions):
            temp = stream.restore()

            # If we set the variable in this scope already
            if var in stream.scope[-1]:

                # we'll check if it's set in one of the older scopes
                for scope in stream.scope[:-1]:
                    if var in scope:
                        # in which case we restore it
                        stream.write('%s = %s' % (var, temp))
                        stream.scope[-1].remove(var)
                        break
                else:
                    stream.write("del %s" % var)

class Condition(object):
    """
      >>> from z3c.pt.generation import CodeIO
      >>> from z3c.pt.testing import pyexp
      >>> from cgi import escape as _escape
      
    Unlimited scope:
    
      >>> stream = CodeIO()
      >>> true = Condition(pyexp("True"))
      >>> false = Condition(pyexp("False"))
      >>> true.begin(stream)
      >>> stream.write("print 'Hello'")
      >>> true.end(stream)
      >>> false.begin(stream)
      >>> stream.write("print 'Universe!'")
      >>> false.end(stream)
      >>> stream.write("print 'World!'")
      >>> exec stream.getvalue()
      Hello
      World!

    Finalized limited scope:

      >>> stream = CodeIO()
      >>> from StringIO import StringIO
      >>> _out = StringIO()
      >>> true = Condition(pyexp("True"), [Write(pyexp("'Hello'"))])
      >>> false = Condition(pyexp("False"), [Write(pyexp("'Hallo'"))])
      >>> true.begin(stream)
      >>> true.end(stream)
      >>> false.begin(stream)
      >>> false.end(stream)
      >>> exec stream.getvalue()
      >>> _out.getvalue()
      'Hello'

    Open limited scope:

      >>> stream = CodeIO()
      >>> from StringIO import StringIO
      >>> _out = StringIO()
      >>> true = Condition(pyexp("True"), [Tag('div')], finalize=False)
      >>> false = Condition(pyexp("False"), [Tag('span')], finalize=False)
      >>> true.begin(stream)
      >>> stream.out("Hello World!")
      >>> true.end(stream)
      >>> false.begin(stream)
      >>> false.end(stream)
      >>> exec stream.getvalue()
      >>> _out.getvalue()
      '<div>Hello World!</div>'
          
    """
      
    def __init__(self, value, clauses=None, finalize=True):
        self.assign = Assign(value)
        self.clauses = clauses
        self.finalize = finalize
        
    def begin(self, stream):
        temp = stream.save()
        self.assign.begin(stream, temp)
        stream.write("if %s:" % temp)
        stream.indent()
        if self.clauses:
            for clause in self.clauses:
                clause.begin(stream)
            if self.finalize:
                for clause in reversed(self.clauses):
                    clause.end(stream)
            stream.outdent()
        
    def end(self, stream):
        temp = stream.restore()

        if self.clauses:
            if not self.finalize:
                stream.write("if %s:" % temp)
                stream.indent()
                for clause in reversed(self.clauses):
                    clause.end(stream)
                    stream.outdent()
        else:
            stream.outdent()
        self.assign.end(stream)

class Else(object):
    def __init__(self, clauses=None):
        self.clauses = clauses
        
    def begin(self, stream):
        stream.write("else:")
        stream.indent()
        if self.clauses:
            for clause in self.clauses:
                clause.begin(stream)
            for clause in reversed(self.clauses):
                clause.end(stream)
            stream.outdent()
        
    def end(self, stream):
        if not self.clauses:
            stream.outdent()

class Group(object):
    def __init__(self, clauses):
        self.clauses = clauses
        
    def begin(self, stream):
        for clause in self.clauses:
            clause.begin(stream)
        for clause in reversed(self.clauses):
            clause.end(stream)
        
    def end(self, stream):
        pass
    
class Tag(object):
    """
      >>> from z3c.pt.generation import CodeIO
      >>> from z3c.pt.testing import pyexp
      >>> from StringIO import StringIO
      >>> from cgi import escape as _escape

      Dynamic attribute:
      
      >>> _out = StringIO(); stream = CodeIO()
      >>> tag = Tag('div', dict(alt=pyexp(repr('Hello World!'))))
      >>> tag.begin(stream)
      >>> stream.out('Hello Universe!')
      >>> tag.end(stream)
      >>> exec stream.getvalue()
      >>> _out.getvalue()
      '<div alt="Hello World!">Hello Universe!</div>'

      Self-closing tag:
      
      >>> _out = StringIO(); stream = CodeIO()
      >>> tag = Tag('br', {}, True)
      >>> tag.begin(stream)
      >>> tag.end(stream)
      >>> exec stream.getvalue()
      >>> _out.getvalue()
      '<br />'

      Unicode:
      
      >>> _out = StringIO(); stream = CodeIO()
      >>> tag = Tag('div', dict(alt=pyexp(repr('La Peña'))))
      >>> tag.begin(stream)
      >>> stream.out('Hello Universe!')
      >>> tag.end(stream)
      >>> exec stream.getvalue()
      >>> _out.getvalue() == '<div alt="La Peña">Hello Universe!</div>'
      True
            
    """

    def __init__(self, tag, attributes={}, selfclosing=False):
        i = tag.find('}')

        if i != -1:
            self.tag = tag[i+1:]
        else:
            self.tag = tag

        self.selfclosing = selfclosing
        self.attributes = attributes
        
    def begin(self, stream):
        stream.out('<%s' % self.tag)

        # static attributes
        static = filter(
            lambda (attribute, value): \
            not isinstance(value, types.expression),
            self.attributes.items())

        dynamic = filter(
            lambda (attribute, value): \
            isinstance(value, types.expression),
            self.attributes.items())

        for attribute, expression in static:
            stream.out(' %s="%s"' %
               (attribute,
                escape(expression, '"')))

        temp = stream.save()

        for attribute, value in dynamic:
            assign = Assign(value)
            assign.begin(stream, temp)
            
            # only include attribute if value is non-trivial
            stream.write("if %s is not None:" % temp)
            stream.indent()

            #if not value.options.get('nocall'):
            #    # if callable, evaluate method
            #    stream.write("if callable(%s): %s = %s()" % (temp, temp, temp))

            if unicode_required_flag:
                stream.write("if isinstance(%s, unicode):" % temp)
                stream.indent()
                stream.write("%s = %s.encode('utf-8')" % (temp, temp))
                stream.outdent()
                stream.write("else:")
                stream.indent()
                stream.write("%s = str(%s)" % (temp, temp))
                stream.outdent()
            else:
                stream.write("%s = str(%s)" % (temp, temp))
                
            stream.write("_out.write(' %s=\"' + _escape(%s, \"\\\"\"))" %
                         (attribute, temp))
            stream.out('"')
            
            assign.end(stream)
            stream.outdent()

        stream.restore()

        if self.selfclosing:
            stream.out(" />")
        else:
            stream.out(">")

    def end(self, stream):
        if not self.selfclosing:
            stream.out('</%s>' % self.tag)

class Repeat(object):
    """
      >>> from z3c.pt.generation import CodeIO
      >>> from z3c.pt.testing import pyexp

    We need to set up the repeat object.

      >>> from z3c.pt import utils
      >>> repeat = utils.repeatdict()

    Simple repeat loop and repeat data structure:

      >>> stream = CodeIO()
      >>> _repeat = Repeat("i", pyexp("range(5)"))
      >>> _repeat.begin(stream)
      >>> stream.write("r = repeat['i']")
      >>> stream.write("print (i, r.index, r.start, r.end, r.number(), r.odd(), r.even())")
      >>> _repeat.end(stream)
      >>> exec stream.getvalue()
      (0, 0, True, False, 1, False, True)
      (1, 1, False, False, 2, True, False)
      (2, 2, False, False, 3, False, True)
      (3, 3, False, False, 4, True, False)
      (4, 4, False, True, 5, False, True)
      >>> _repeat.end(stream)

    A repeat over an empty set.
    
      >>> stream = CodeIO()
      >>> _repeat = Repeat("j", pyexp("range(0)"))
      >>> _repeat.begin(stream)
      >>> _repeat.end(stream)
      >>> exec stream.getvalue()

    """
        
    def __init__(self, v, e, scope=()):
        self.variable = v
        self.define = Define(v, types.value("None"))
        self.assign = Assign(e)

    def begin(self, stream):
        variable = self.variable
        iterator = stream.save()

        # assign iterator
        self.assign.begin(stream, iterator)

        # initialize variable scope
        self.define.begin(stream)

        # initialize iterator
        stream.write("%s = repeat.insert('%s', %s)" % (iterator, variable, iterator))
        
        # loop
        stream.write("try:")
        stream.indent()
        stream.write("while True:")
        stream.indent()
        stream.write("%s = %s.next()" % (variable, iterator))
        
    def end(self, stream):
        # cook before leaving loop
        stream.cook()
        
        stream.outdent()
        stream.outdent()
        stream.write("except StopIteration:")
        stream.indent()
        stream.write("pass")
        stream.outdent()
        
        self.define.end(stream)
        self.assign.end(stream)
        stream.restore()

class Write(object):
    """
    >>> from z3c.pt.generation import CodeIO; stream = CodeIO()
    >>> from z3c.pt.testing import pyexp
    >>> from StringIO import StringIO
    >>> from cgi import escape as _escape

    Basic write:
    
    >>> _out = StringIO()
    >>> write = Write(pyexp("'New York'"))
    >>> write.begin(stream)
    >>> write.end(stream)
    >>> exec stream.getvalue()
    >>> _out.getvalue()
    'New York'

    Try-except parts:

    >>> stream = CodeIO()
    >>> _out = StringIO()
    >>> write = Write(pyexp("undefined | 'New Delhi'"))
    >>> write.begin(stream)
    >>> write.end(stream)
    >>> exec stream.getvalue()
    >>> _out.getvalue()
    'New Delhi'

    Unicode:

    >>> stream = CodeIO()
    >>> _out = StringIO()
    >>> write = Write(types.value("unicode('La Pe\xc3\xb1a', 'utf-8')"))
    >>> write.begin(stream)
    >>> write.end(stream)
    >>> exec stream.getvalue()
    >>> _out.getvalue() == 'La Pe\xc3\xb1a'
    True
    
    """

    value = assign = None
    
    def __init__(self, value):
        if isinstance(value, types.parts):
            self.assign = Assign(value)
        else:
            self.value = value

        self.structure = not isinstance(value, types.escape)
        
    def begin(self, stream):
        temp = stream.save()

        if self.value:
            expr = self.value
        else:
            self.assign.begin(stream, temp)
            expr = temp

        stream.write("_urf = %s" % expr)
        stream.write("if _urf is None: _urf = ''")

        if unicode_required_flag:
            stream.write("if isinstance(_urf, unicode):")
            stream.indent()
            stream.write("_out.write(_urf.encode('utf-8'))")
            stream.outdent()
            stream.write("else:")
            stream.indent()
            if self.structure:
                stream.write("_out.write(str(_urf))")
            else:
                stream.write("_out.write(_escape(str(_urf)))")
            stream.outdent()
        else:
            if self.structure:
                stream.write("_out.write(str(_urf))")
            else:
                stream.write("_out.write(_escape(str(_urf)))")
            
    def end(self, stream):
        if self.assign:
            self.assign.end(stream)
        stream.restore()

class Out(object):
    """
      >>> from z3c.pt.generation import CodeIO; stream = CodeIO()
      >>> from z3c.pt.testing import pyexp
      >>> from StringIO import StringIO
      >>> _out = StringIO()
      
      >>> out = Out('Hello World!')
      >>> out.begin(stream)
      >>> out.end(stream)
      >>> exec stream.getvalue()
      >>> _out.getvalue()
      'Hello World!'      
    """
    
    def __init__(self, string, defer=False):
        self.string = string
        self.defer = defer
        
    def begin(self, stream):
        if not self.defer:
            stream.out(self.string)

    def end(self, stream):
        if self.defer:
            stream.out(self.string)

class Translate(object):
    """
    The translate clause works retrospectively.
    """

    def begin(self, stream):
        raise

    def end(self, stream):
        raise



import zope.i18n

import cgi
import StringIO
import cStringIO

import expressions
import utils

import z3c.pt.generation

wrapper = """\
def render(%starget_language=None):
\tglobal generation

\t_out = generation.initialize_stream()
    
\t(_attributes, repeat) = generation.initialize_tal()
\t(_domain, _translate) = generation.initialize_i18n()
\t(_escape, _marker) = generation.initialize_helpers()
\t_path = generation.initialize_traversal()

\t_target_language = target_language
%s
\treturn _out.getvalue().decode('utf-8')
"""

def initialize_i18n():
    return (None, zope.i18n.translate)

def initialize_tal():
    return ({}, utils.repeatdict())

def initialize_helpers():
    return (cgi.escape, object())

def initialize_stream():
    return cStringIO.StringIO()

def initialize_traversal():
    return expressions.PathTranslation.traverse

class Generator(object):
    def __init__(self, params):
        self.params = tuple(params)
        self.stream = CodeIO(indentation=1, indentation_string="\t")

        # initialize variable scope
        self.stream.scope.append(set(params + ['_out']))

    def __call__(self):
        # prepare template arguments
        args = ', '.join(self.params)
        if args: args += ', '

        code = self.stream.getvalue()

        return wrapper % (args, code), {'generation': z3c.pt.generation}
        
class CodeIO(StringIO.StringIO):
    """A I/O stream class that provides facilities to generate Python code.

    * Indentation is managed using ``indent`` and ``outdent``.

    * Annotations can be assigned on a per-line basis using ``annotate``.

    * Convenience methods for keeping track of temporary variables
   
    * Methods to process clauses (see ``begin`` and ``end``).
    
    """

    t_prefix = '_tmp'
    v_prefix = '_var'

    def __init__(self, indentation=0, indentation_string="\t"):
        StringIO.StringIO.__init__(self)
        self.indentation = indentation
        self.indentation_string = indentation_string
        self.queue = u''
        self.scope = [set()]
        self.annotations = {}
        
        self._variables = {}
        self.t_counter = 0
        self.l_counter = 0
        
    def save(self):
        self.t_counter += 1
        return "%s%d" % (self.t_prefix, self.t_counter)
        
    def restore(self):
        var = "%s%d" % (self.t_prefix, self.t_counter)
        self.t_counter -= 1
        return var
        
    def indent(self, amount=1):
        if amount > 0:
            self.cook()
            self.indentation += amount

    def outdent(self, amount=1):
        if amount > 0:
            self.cook()
            self.indentation -= amount

    def annotate(self, item):
        self.annotations[self.l_counter] = item

    def out(self, string):
        self.queue += string
        
    def cook(self):
        if self.queue:
            queue = self.queue
            self.queue = ''
            self.write("_out.write('%s')" %
                       queue.replace('\n', '\\n').replace("'", "\\'"))
            
    def write(self, string):
        if isinstance(string, str):
            string = string.decode('utf-8')

        self.l_counter += len(string.split('\n'))-1
        
        self.cook()
        StringIO.StringIO.write(
            self, self.indentation_string * self.indentation + string + '\n')

    def getvalue(self):
        self.cook()
        return StringIO.StringIO.getvalue(self)
            
    def begin(self, clauses):
        if isinstance(clauses, (list, tuple)):
            for clause in clauses:
                self.begin(clause)
        else:
            clauses.begin(self)
            
    def end(self, clauses):
        if isinstance(clauses, (list, tuple)):
            for clause in reversed(clauses):
                self.end(clause)
        else:
            clauses.end(self)
        



from compiler import ast, parse
from compiler.pycodegen import ModuleCodeGenerator

from transformer import ASTTransformer

marker = object()

CONSTANTS = frozenset(['False', 'True', 'None', 'NotImplemented', 'Ellipsis'])

UNDEFINED = object()

class Lookup(object):
    """Abstract base class for variable lookup implementations."""

    def globals(cls):
        """Construct the globals dictionary to use as the execution context for
        the expression or suite.
        """
        return {
            '_lookup_attr': cls.lookup_attr,
        }

    globals = classmethod(globals)

    @classmethod
    def lookup_attr(cls, obj, key):
        try:
            return getattr(obj, key)
        except AttributeError, e:
            try:
                return obj[key]
            except (KeyError, TypeError):
                raise e

class TemplateASTTransformer(ASTTransformer):
    def __init__(self):
        self.locals = [CONSTANTS]
        
    def visitGetattr(self, node):
        """
        Allow fallback to dictionary lookup if attribute does not exist.

        Variables starting with an underscore are exempt.

        """
        
        if hasattr(node.expr, 'name') and node.expr.name.startswith('_'):
            return ast.Getattr(node.expr, node.attrname)

        return ast.CallFunc(ast.Name('_lookup_attr'), [
            self.visit(node.expr), ast.Const(node.attrname)
            ])
    
class Suite(object):
    __slots__ = ['code', '_globals']

    xform = TemplateASTTransformer
    mode = 'exec'
    
    def __init__(self, source):
        """Create the code object from a string."""

        node = parse(source, self.mode)

        # build tree
        transform = self.xform()
        tree = transform.visit(node)
        filename = tree.filename = '<script>'

        # generate code
        gen = ModuleCodeGenerator(tree)
        gen.optimized = True

        self._globals = Lookup.globals()
        self.code = gen.getCode()
        
    def __hash__(self):
        return hash(self.code)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.source)



import template
import translation

class TextTemplate(template.BaseTemplate):
    @property
    def translate(self):
        return translation.translate_text

class TextTemplateFile(template.BaseTemplateFile):
    @property
    def translate(self):
        return translation.translate_text

class ViewTextTemplate(property):
    def __init__(self, body):
        self.template = TextTemplate(body)
        property.__init__(self, self.render)

    def render(self, view):
        def template(**kwargs):
            return self.template.render(view=view,
                                        context=view.context,
                                        request=view.request,
                                        options=kwargs)
        return template        
    
class ViewTextTemplateFile(ViewTextTemplate):
    def __init__(self, filename):
        self.template = TextTemplateFile(filename)
        property.__init__(self, self.render)



from zope import interface

class IExpressionTranslation(interface.Interface):
    """This interface defines an expression translation utility."""
    
    def name(string):
        """Should interpret ``string`` as a name. This method will
        usually be the identity function."""
    
    def value(string):
        """Translate ``string`` to a value-expression tuple.

        Specification:

           value :: = expression [ |* value ]
           expresion ::= an expression string

           *) Using | as _logical or_ is not supported.

        """

    def validate(string):
        """Raises exception if ``string`` is not a valid exception."""

    def translate(string):
        """Translates ``string``."""
        
    def search(string):
        """Extracts the longest valid expression from the beginning of
        the provided string."""

    def variable(string):
        """A variable definition.
        
        Specification:
        
           variables :: = variable_name [, variables]

        """
        
    def mapping(string):
        """A mapping definition."""

    def definition(string):
        """A definition."""

    def definitions(string):
        """Multiple definitions.
        
        Specification:

           argument ::= define_var [';' define_var]
           define_var ::= Name python_expression
           
        """



from zope import component

from StringIO import StringIO
import lxml.etree

import generation
import expressions
import clauses
import interfaces
import utils
import types

def attribute(ns, factory=None, default=None):
    def get(self):
        value = self.attrib.get(ns)
        if value is not None:
            if factory is None:
                return value

            f = factory(self._translator())
            return f(value)
        elif default is not None:
            return default
        
    def set(self, value):
        self.attrib[ns] = value

    return property(get, set)

class Element(lxml.etree.ElementBase):
    def begin(self, stream):
        stream.scope.append(set())
        stream.begin(self._clauses())
        
    def end(self, stream):
        stream.end(self._clauses())
        stream.scope.pop()

    def body(self, stream):
        skip = self.replace or self.content or self.i18n_translate is not None
        if not skip:
            for element in [e for e in self if isinstance(e, lxml.etree._Comment)]:
                self._wrap_comment(element)
            for element in self:
                element.interpolate(stream)
            for element in self:
                element.visit(stream)
                    
    def visit(self, stream):
        self.begin(stream)
        self.body(stream)
        self.end(stream)

    def interpolate(self, stream):
        """The current interpolation strategy is to translate the
        interpolation statements into TAL."""
        
        translator = self._translator()
        if self.text is not None:
            while self.text:
                m = translator.interpolate(self.text)
                if m is None:
                    break

                t = parser.makeelement(
                    '{http://xml.zope.org/namespaces/tal}interpolation')
                t.attrib['replace'] = m.group('expression')
                t.tail = self.text[m.end():]
                self.insert(0, t)                
                self.text = self.text[:m.start()+1]

        # interpolate tail
        if self.tail is not None:
            while self.tail:
                m = translator.interpolate(self.tail)
                if m is None:
                    break

                t = parser.makeelement(
                    '{http://xml.zope.org/namespaces/tal}interpolation')
                t.attrib['replace'] = m.group('expression')
                t.tail = self.tail[m.end():]
                parent = self.getparent()
                parent.insert(parent.index(self)+1, t)                
                self.tail = self.tail[:m.start()+1]

        # interpolate attributes
        for name in self._static_attributes():
            value = self.attrib[name]

            if translator.interpolate(value):
                del self.attrib[name]

                attributes = '{http://xml.zope.org/namespaces/tal}attributes'
                if attributes in self.attrib:
                    self.attrib[attributes] += '; %s string: %s' % (name, value)
                else:
                    self.attrib[attributes] = '%s string: %s' % (name, value)
                
    def _clauses(self):
        _ = []

        # i18n domain
        if self.i18n_domain is not None:
            _.append(clauses.Define(
                "_domain", types.value(repr(self.i18n_domain))))

        # defines
        if self.define is not None:
            for variables, expression in self.define:
                _.append(clauses.Define(variables, expression))

        # condition
        if self.condition is not None:
            _.append(clauses.Condition(self.condition))

        # repeat
        if self.repeat is not None:
            variables, expression = self.repeat
            if len(variables) != 1:
                raise ValueError, "Cannot unpack more than one variable in a repeat statement."
            _.append(clauses.Repeat(variables[0], expression))

        # tag tail (deferred)
        if self.tail:
            _.append(clauses.Out(self.tail, defer=True))

        # compute dynamic flag
        dynamic = self.replace or self.content or self.i18n_translate is not None
        
        # tag
        if self.replace is None:
            selfclosing = self.text is None and not dynamic and len(self) == 0
            tag = clauses.Tag(self.tag, self._attributes(), selfclosing=selfclosing)

            if self.omit:
                _.append(clauses.Condition(_not(self.omit), [tag], finalize=False))
            elif self.omit is not None:
                pass
            else:
                _.append(tag)

        # tag text (if we're not replacing tag body)
        if self.text and not dynamic:
            _.append(clauses.Out(self.text))

        # dynamic content and content translation
        replace = self.replace
        content = self.content

        if replace and content:
            raise ValueError, "Can't use replace clause together with content clause."

        expression = replace or content
        if expression:
            if self.i18n_translate is not None:
                if self.i18n_translate != "":
                    raise ValueError, "Can't use message id with dynamic content translation."
                _.append(clauses.Translate())

            _.append(clauses.Write(expression))
        else:
            if self.i18n_translate is not None:
                msgid = self.i18n_translate
                if not msgid:
                    msgid = self._msgid()

                # for each named block, create a new output stream
                # and use the value in the translation mapping dict

                elements = [e for e in self if e.i18n_name]

                if elements:
                    mapping = '_mapping'
                    _.append(clauses.Assign(types.value('{}'), mapping))
                else:
                    mapping = 'None'
                    
                for element in elements:
                    name = element.i18n_name
                    
                    subclauses = []
                    subclauses.append(clauses.Define(
                        '_out', types.value('generation.initialize_stream()')))
                    subclauses.append(clauses.Group(element._clauses()))
                    subclauses.append(clauses.Assign(
                        types.value('_out.getvalue()'), "%s['%s']" % (mapping, name)))

                    _.append(clauses.Group(subclauses))

                _.append(clauses.Assign(
                    _translate(types.value(repr(msgid)), mapping=mapping, default='_marker'),
                    '_result'))

                # write translation to output if successful, otherwise
                # fallback to default rendition; 
                result = types.value('_result')
                condition = types.value('_result is not _marker')
                _.append(clauses.Condition(condition, [clauses.Write(result)]))

                subclauses = []
                if self.text:
                    subclauses.append(clauses.Out(self.text))
                for element in self:
                    name = element.i18n_name
                    if name:
                        value = types.value("%s['%s']" % (mapping, name))
                        subclauses.append(clauses.Write(value))
                    else:
                        subclauses.append(clauses.Out(lxml.etree.tostring(element)))

                if subclauses:
                    _.append(clauses.Else(subclauses))

        return _

    def _wrap_comment(self, element):
        index = self.index(element)

        t = parser.makeelement(
            '{http://xml.zope.org/namespaces/tal}comment')

        t.attrib['omit-tag'] = ''
        t.tail = element.tail
        t.text = '<!--' + element.text + '-->'

        for child in element.getchildren():
            t.append(child)

        self.remove(element)
        self.insert(index, t)
    
    def _msgid(self):
        """Create an i18n msgid from the tag contents."""

        out = StringIO(self.text)
        for element in self:
            name = element.i18n_name
            if name:
                out.write("${%s}" % name)
                out.write(element.tail)
            else:
                out.write(lxml.etree.tostring(element))

        msgid = out.getvalue().strip()
        msgid = msgid.replace('  ', ' ').replace('\n', '')
        
        return msgid

    def _static_attributes(self):
        attributes = {}

        for key in self.keys():
            if not key.startswith('{'):
                attributes[key] = self.attrib[key]

        return attributes
        
    def _attributes(self):
        """Aggregate static, dynamic and translatable attributes."""

        # static attributes are at the bottom of the food chain
        attributes = self._static_attributes()
        
        # dynamic attributes
        attrs = self.attributes
        if attrs is not None:
            for variables, expression in attrs:
                if len(variables) != 1:
                    raise ValueError, "Tuple definitions in assignment clause is not supported."

                variable = variables[0]
                attributes[variable] = expression
        else:
            attrs = []

        dynamic = [key for (key, expression) in attrs]

        # translated attributes
        if self.i18n_attributes:
            for variable, msgid in self.i18n_attributes:
                if msgid:
                    if variable in dynamic:
                        raise ValueError, "Message id not allowed in conjunction with " + \
                                          "a dynamic attribute."

                    value = types.value(msgid)

                    if variable in static:
                        expression = _translate(value, default=attributes[variable])
                    else:
                        expression = _translate(value)
                else:
                    if variable in dynamic or variable in static:
                        expression = _translate(attributes[variable])
                    else:
                        raise ValueError, "Must be either static or dynamic attribute " + \
                                          "when no message id is supplied."

                attributes[variable] = expression

        return attributes

    def _translator(self):
        while self.default_expression is None:
            self = self.getparent()
            if self is None:
                raise ValueError("Default expression not set.")
            
        return component.getUtility(
            interfaces.IExpressionTranslation, name=self.default_expression)

    define = attribute(
        "{http://xml.zope.org/namespaces/tal}define", lambda p: p.definitions)
    condition = attribute(
        "{http://xml.zope.org/namespaces/tal}condition", lambda p: p.expression)
    repeat = attribute(
        "{http://xml.zope.org/namespaces/tal}repeat", lambda p: p.definition)
    attributes = attribute(
        "{http://xml.zope.org/namespaces/tal}attributes", lambda p: p.definitions)
    content = attribute(
        "{http://xml.zope.org/namespaces/tal}content", lambda p: p.output)
    replace = attribute(
        "{http://xml.zope.org/namespaces/tal}replace", lambda p: p.output)
    omit = attribute(
        "{http://xml.zope.org/namespaces/tal}omit-tag", lambda p: p.expression)
    i18n_translate = attribute(
        "{http://xml.zope.org/namespaces/i18n}translate")
    i18n_attributes = attribute(
        "{http://xml.zope.org/namespaces/i18n}attributes", lambda p: p.mapping)
    i18n_domain = attribute(
        "{http://xml.zope.org/namespaces/i18n}domain")
    i18n_name = attribute(
        "{http://xml.zope.org/namespaces/i18n}name")
    default_expression = attribute(
        "{http://xml.zope.org/namespaces/tal}default-expression")
    
class TALElement(Element):
    define = attribute("define", lambda p: p.definitions)
    replace = attribute("replace", lambda p: p.output)
    repeat = attribute("repeat", lambda p: p.definition)
    attributes = attribute("attributes", lambda p: p.expression)
    content = attribute("content", lambda p: p.output)
    omit = attribute("omit-tag", lambda p: p.expression, u"")
    default_expression = attribute("default-expression", lambda p: p.name)
    
    def _static_attributes(self):
        attributes = {}

        for key in self.keys():
            if key not in \
                   ('define',
                    'replace',
                    'repeat',
                    'attributes',
                    'content',
                    'omit-tag'):
                raise ValueError(
                    u"Attribute '%s' not allowed in the namespace '%s'" %
                    (key, self.nsmap[self.prefix]))

        return attributes

# set up namespaces for XML parsing
lookup = lxml.etree.ElementNamespaceClassLookup()
parser = lxml.etree.XMLParser()
parser.setElementClassLookup(lookup)

try:
    lookup.get_namespace('http://www.w3.org/1999/xhtml')[None] = Element
    lookup.get_namespace('http://xml.zope.org/namespaces/tal')[None] = TALElement
except AttributeError:
    lxml.etree.Namespace('http://www.w3.org/1999/xhtml')[None] = Element
    lxml.etree.Namespace('http://xml.zope.org/namespaces/tal')[None] = TALElement

def translate_xml(body, *args, **kwargs):
    tree = lxml.etree.parse(StringIO(body), parser)
    root = tree.getroot()
    return translate_etree(root, *args, **kwargs)

def translate_etree(root, params=[], default_expression='python'):
    if None not in root.nsmap:
        raise ValueError, "Must set default namespace."

    # set default expression name
    key = '{http://xml.zope.org/namespaces/tal}default-expression'
    if key not in root.attrib:
        root.attrib[key] = default_expression

    # set up code generation stream
    generator = generation.Generator(params)
    stream = generator.stream

    # visit root
    root.interpolate(stream)
    root.visit(stream)

    return generator

def translate_text(body, *args, **kwargs):
    xml = parser.makeelement(
        '{http://www.w3.org/1999/xhtml}text',
        nsmap={None: 'http://www.w3.org/1999/xhtml'})
    xml.text = body
    xml.attrib['{http://xml.zope.org/namespaces/tal}omit-tag'] = ''
    return translate_etree(xml, *args, **kwargs)
    
def _translate(value, mapping=None, default=None):
    format = "_translate(%s, domain=_domain, mapping=%s, "\
             "target_language=_target_language, default=%s)"
    
    return types.value(format % (value, mapping, default))

def _not(value):
    return types.value("not (%s)" % value)



class expression:
    pass

class parts(tuple, expression):
    def __repr__(self):
        return 'parts'+tuple.__repr__(self)

class value(str, expression):            
    def __repr__(self):
        return 'value(%s)' % str.__repr__(self)

class join(tuple, expression):
    def __repr__(self):
        return 'join'+tuple.__repr__(self)

class declaration(tuple):
    def __repr__(self):
        return 'declaration'+tuple.__repr__(self)

class mapping(tuple):
    def __repr__(self):
        return 'mapping'+tuple.__repr__(self)

class definitions(tuple):
    def __repr__(self):
        return 'definitions'+tuple.__repr__(self)

class escape(parts):
    def __repr__(self):
        return 'escape'+tuple.__repr__(self)



import os
import sys
import codegen
import traceback

class BaseTemplate(object):
    registry = {}
    default_expression = 'python'
    
    def __init__(self, body, default_expression=None):
        self.body = body
        self.signature = hash(body)        

        if default_expression:
            self.default_expression = default_expression
            
    @property
    def translate(self):
        return NotImplementedError("Must be implemented by subclass.")

    def cook(self, params):
        generator = self.translate(
            self.body, params=params, default_expression=self.default_expression)
        
        source, _globals = generator()
         
        suite = codegen.Suite(source)
        
        self.source = source
        self.annotations = generator.stream.annotations
        
        _globals.update(suite._globals)
        _locals = {}

        exec suite.code in _globals, _locals

        return _locals['render']

    def render(self, **kwargs):
        signature = self.signature + hash(",".join(kwargs.keys()))

        template = self.registry.get(signature)
        if not template:
            self.registry[signature] = template = self.cook(kwargs.keys())

        try:
            return template(**kwargs)
        except Exception, e:
            __traceback_info__ = getattr(e, '__traceback_info__', None)
            if __traceback_info__ is not None:
                raise e
            
            etype, value, tb = sys.exc_info()
            lineno = tb.tb_next.tb_lineno-1
            annotations = self.annotations

            while lineno >= 0:
                if lineno in annotations:
                    annotation = annotations.get(lineno)
                    break

                lineno -= 1
            else:
                annotation = "n/a"

            e.__traceback_info__ = "While rendering %s, an exception was "\
                                   "raised evaluating ``%s``:\n\n" % \
                                   (repr(self), str(annotation))
            
            e.__traceback_info__ += "".join(traceback.format_tb(tb))
            
            raise e            
        
    def __call__(self, **kwargs):
        return self.render(**kwargs)

    def __repr__(self):
        return u"<%s %d>" % (self.__class__.__name__, id(self))

class BaseTemplateFile(BaseTemplate):
    def __init__(self, filename):
        if not os.path.isabs(filename):
            package_name = sys._getframe(2).f_globals['__name__']
            module = sys.modules[package_name]
            try:
                path = module.__path__[0]
            except AttributeError:
                path = module.__file__
                path = path[:path.rfind(os.sep)]
                
            filename = path + os.sep + filename

        # make sure file exists
        os.lstat(filename)
        self.filename = filename
                
    def _get_filename(self):
        return getattr(self, '_filename', None)

    def _set_filename(self, filename):
        self._filename = filename
        self._v_last_read = False

    filename = property(_get_filename, _set_filename)

    def render(self, **kwargs):
        if self._cook_check():
            self.body = open(self.filename, 'r').read()
            self.signature = hash(self.body)
            self._v_last_read = self.mtime()

        return BaseTemplate.render(self, **kwargs)
            
    def _cook_check(self):
        if self._v_last_read and not __debug__:
            return

        if self.mtime() == self._v_last_read:
            return

        return True

    def mtime(self):
        try:
            return os.path.getmtime(self.filename)
        except OSError:
            return 0

    def __repr__(self):
        return u"<%s %s>" % (self.__class__.__name__, self.filename)
